from typing import Any

from fastapi import status
from langchain.chains.base import Chain
from starlette.types import Send

from lanarky.events import Events, ServerSentEvent, ensure_bytes
from lanarky.logging import logger
from lanarky.responses import HTTPStatusDetail
from lanarky.responses import StreamingResponse as _StreamingResponse


class StreamingResponse(_StreamingResponse):
    """StreamingResponse class for LangChain resources."""

    def __init__(
        self,
        chain: Chain,
        config: dict[str, Any],
        *args: Any,
        **kwargs: dict[str, Any],
    ) -> None:
        """Constructor method.

        Args:
            chain: A LangChain instance.
            config: A config dict.
            *args: Positional arguments to pass to the parent constructor.
            **kwargs: Keyword arguments to pass to the parent constructor.
        """
        super().__init__(*args, **kwargs)

        self.chain = chain
        self.config = config

    async def stream_response(self, send: Send) -> None:
        """Stream LangChain outputs.

        If an exception occurs while iterating over the LangChain, an
        internal server error is sent to the client.

        Args:
            send: The ASGI send callable.
        """
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )

        if "callbacks" in self.config:
            for callback in self.config["callbacks"]:
                if hasattr(callback, "send"):
                    callback.send = send

        try:
            # TODO: migrate to `.ainvoke` when adding support
            # for LCEL
            outputs = await self.chain.acall(**self.config)
            if self.background is not None:
                self.background.kwargs.update({"outputs": outputs})
        except Exception as e:
            logger.error(f"chain runtime error: {e}")
            if self.background is not None:
                self.background.kwargs.update({"outputs": {}, "error": e})
            chunk = ServerSentEvent(
                data=dict(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=HTTPStatusDetail.INTERNAL_SERVER_ERROR,
                ),
                event=Events.ERROR,
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": ensure_bytes(chunk, None),
                    "more_body": True,
                }
            )

        await send({"type": "http.response.body", "body": b"", "more_body": False})



from typing import Any

from fastapi.applications import AppType, FastAPI
from fastapi.openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.requests import Request
from fastapi.responses import HTMLResponse, JSONResponse


class Lanarky(FastAPI):
    """The main application class.

    To know more about the FastAPI parameters, read the
    [FastAPI documentation](https://fastapi.tiangolo.com/reference/fastapi/).
    """

    def __init__(
        self: AppType, *, title: str = "Lanarky", **kwargs: dict[str, Any]
    ) -> None:
        """Constructor method.

        Args:
            title: The title of the application.
            **kwargs: Additional arguments to pass to the FastAPI constructor.
        """
        super().__init__(title=title, **kwargs)

    def setup(self) -> None:  # pragma: no cover
        """Setup the application.

        Overrides the `setup` method of the FastAPI class.
        """
        if self.openapi_url:
            urls = (server_data.get("url") for server_data in self.servers)
            server_urls = {url for url in urls if url}

            async def openapi(req: Request) -> JSONResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                if root_path not in server_urls:
                    if root_path and self.root_path_in_servers:
                        self.servers.insert(0, {"url": root_path})
                        server_urls.add(root_path)
                return JSONResponse(self.openapi())

            self.add_route(self.openapi_url, openapi, include_in_schema=False)
        if self.openapi_url and self.docs_url:

            async def swagger_ui_html(req: Request) -> HTMLResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                oauth2_redirect_url = self.swagger_ui_oauth2_redirect_url
                if oauth2_redirect_url:
                    oauth2_redirect_url = root_path + oauth2_redirect_url
                return get_swagger_ui_html(
                    openapi_url=openapi_url,
                    title=self.title + " - Swagger UI",
                    oauth2_redirect_url=oauth2_redirect_url,
                    init_oauth=self.swagger_ui_init_oauth,
                    swagger_ui_parameters=self.swagger_ui_parameters,
                    swagger_favicon_url="https://lanarky.ajndkr.com/assets/icon.svg",
                )

            self.add_route(self.docs_url, swagger_ui_html, include_in_schema=False)

            if self.swagger_ui_oauth2_redirect_url:

                async def swagger_ui_redirect(req: Request) -> HTMLResponse:
                    return get_swagger_ui_oauth2_redirect_html()

                self.add_route(
                    self.swagger_ui_oauth2_redirect_url,
                    swagger_ui_redirect,
                    include_in_schema=False,
                )
        if self.openapi_url and self.redoc_url:

            async def redoc_html(req: Request) -> HTMLResponse:
                root_path = req.scope.get("root_path", "").rstrip("/")
                openapi_url = root_path + self.openapi_url
                return get_redoc_html(
                    openapi_url=openapi_url,
                    title=self.title + " - ReDoc",
                    redoc_favicon_url="https://lanarky.ajndkr.com/assets/icon.svg",
                )

            self.add_route(self.redoc_url, redoc_html, include_in_schema=False)



import inspect
from typing import Any, Callable, Optional, Sequence

from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute, APIRouter, APIWebSocketRoute

from .utils import build_factory_api_endpoint, build_factory_websocket_endpoint


class LangchainAPIRoute(APIRoute):
    """APIRoute class for LangChain."""

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = Default(None),
        **kwargs: dict[str, Any],
    ) -> None:
        """Constructor method.

        Args:
            path: The path for the route.
            endpoint: The endpoint to call when the route is requested.
            response_model: The response model to use for the route.
            **kwargs: Keyword arguments to pass to the parent constructor.
        """
        # NOTE: LangchainAPIRoute is initialised again when
        # router is included in app. This is a hack to
        # build the factory endpoint only once.
        if not inspect.iscoroutinefunction(endpoint):
            factory_endpoint = build_factory_api_endpoint(path, endpoint)
            super().__init__(
                path, factory_endpoint, response_model=response_model, **kwargs
            )
        else:
            super().__init__(path, endpoint, response_model=response_model, **kwargs)


class LangchainAPIWebSocketRoute(APIWebSocketRoute):
    """APIWebSocketRoute class for LangChain."""

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        name: Optional[str] = None,
        **kwargs: dict[str, Any],
    ) -> None:
        """Constructor method.

        Args:
            path: The path for the route.
            endpoint: The endpoint to call when the route is requested.
            name: The name of the route.
            **kwargs: Keyword arguments to pass to the parent constructor.
        """
        super().__init__(path, endpoint, name=name, **kwargs)
        # NOTE: LangchainAPIRoute is initialised again when
        # router is included in app. This is a hack to
        # build the factory endpoint only once.
        if not inspect.iscoroutinefunction(endpoint):
            factory_endpoint = build_factory_websocket_endpoint(path, endpoint)
            super().__init__(path, factory_endpoint, name=name, **kwargs)
        else:
            super().__init__(path, endpoint, name=name, **kwargs)


class LangchainAPIRouter(APIRouter):
    """APIRouter class for LangChain."""

    def __init__(self, *, route_class: type[APIRoute] = LangchainAPIRoute, **kwargs):
        super().__init__(route_class=route_class, **kwargs)

    def add_api_websocket_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        name: Optional[str] = None,
        *,
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> None:
        current_dependencies = self.dependencies.copy()
        if dependencies:
            current_dependencies.extend(dependencies)

        route = LangchainAPIWebSocketRoute(
            self.prefix + path,
            endpoint=endpoint,
            name=name,
            dependencies=current_dependencies,
            dependency_overrides_provider=self.dependency_overrides_provider,
        )
        self.routes.append(route)



import re
from typing import Any, Awaitable, Callable

from fastapi import Depends
from langchain.agents import AgentExecutor
from langchain.chains.base import Chain
from langchain.schema.document import Document
from pydantic import BaseModel, create_model
from starlette.routing import compile_path

from lanarky.adapters.langchain.callbacks import (
    FinalTokenStreamingCallbackHandler,
    FinalTokenWebSocketCallbackHandler,
    SourceDocumentsStreamingCallbackHandler,
    SourceDocumentsWebSocketCallbackHandler,
    TokenStreamingCallbackHandler,
    TokenWebSocketCallbackHandler,
)
from lanarky.adapters.langchain.responses import HTTPStatusDetail, StreamingResponse
from lanarky.events import Events
from lanarky.logging import logger
from lanarky.utils import model_dump
from lanarky.websockets import WebSocket, WebsocketSession


def build_factory_api_endpoint(
    path: str, endpoint: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """Build a factory endpoint for API routes.

    Args:
        path: The path for the route.
        endpoint: LangChain instance factory function.
    """
    chain = compile_chain_factory(endpoint)

    # index 1 of `compile_path` contains path_format output
    model_prefix = compile_model_prefix(compile_path(path)[1], chain)
    request_model = create_request_model(chain, model_prefix)

    callbacks = get_streaming_callbacks(chain)

    async def factory_endpoint(
        request: request_model, chain: Chain = Depends(endpoint)
    ):
        return StreamingResponse(
            chain=chain, config={"inputs": model_dump(request), "callbacks": callbacks}
        )

    return factory_endpoint


def build_factory_websocket_endpoint(
    path: str, endpoint: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """Build a factory endpoint for WebSocket routes.

    Args:
        path: The path for the route.
        endpoint: LangChain instance factory function.
    """
    chain = compile_chain_factory(endpoint)

    # index 1 of `compile_path` contains path_format output
    model_prefix = compile_model_prefix(compile_path(path)[1], chain)
    request_model = create_request_model(chain, model_prefix)

    async def factory_endpoint(websocket: WebSocket, chain: Chain = Depends(endpoint)):
        callbacks = get_websocket_callbacks(chain, websocket)
        async with WebsocketSession().connect(websocket) as session:
            async for data in session:
                try:
                    await chain.acall(
                        inputs=model_dump(request_model(**data)),
                        callbacks=callbacks,
                    )
                except Exception as e:
                    logger.error(f"langchain error: {e}")
                    await websocket.send_json(
                        dict(
                            data=dict(
                                status=500,
                                detail=HTTPStatusDetail(
                                    code=500,
                                    message="Internal Server Error",
                                ),
                            ),
                            event=Events.ERROR,
                        )
                    )
                await websocket.send_json(dict(data="", event=Events.END))

    return factory_endpoint


def compile_chain_factory(endpoint: Callable[..., Any]):
    """Compile a LangChain instance factory function.

    Args:
        endpoint: LangChain instance factory function.
    """
    try:
        chain = endpoint()
    except TypeError:
        raise TypeError("set default values for all factory endpoint parameters")

    if not isinstance(chain, Chain):
        raise TypeError("factory endpoint must return a Chain instance")
    return chain


def create_request_model(chain: Chain, prefix: str = "") -> BaseModel:
    """Create a pydantic request model for a LangChain instance.

    Args:
        chain: A LangChain instance.
        prefix: A prefix for the model name.
    """
    request_fields = {}

    for key in chain.input_keys:
        # TODO: add support for other input key types
        # based on demand
        if key == "chat_history":
            request_fields[key] = (list[tuple[str, str]], ...)
        else:
            request_fields[key] = (str, ...)

    prefix = prefix or chain.__class__.__name__

    return create_model(f"{prefix}Request", **request_fields)


def create_response_model(chain: Chain, prefix: str = None) -> BaseModel:
    """Create a pydantic response model for a LangChain instance.

    Args:
        chain: A LangChain instance.
        prefix: A prefix for the model name.
    """
    response_fields = {}

    for key in chain.output_keys:
        # TODO: add support for other output key types
        # based on demand
        if key == "source_documents":
            response_fields[key] = (list[Document], ...)
        else:
            response_fields[key] = (str, ...)

    prefix = prefix or chain.__class__.__name__

    return create_model(f"{prefix}Response", **response_fields)


def compile_model_prefix(path: str, chain: Chain) -> str:
    """Compile a prefix for pydantic models.

    Args:
        path: The path for the route.
        chain: A LangChain instance.
    """
    # Remove placeholders like '{item}' using regex
    path_wo_params = re.sub(r"\{.*?\}", "", path)
    path_prefix = "".join([part.capitalize() for part in path_wo_params.split("/")])

    chain_prefix = chain.__class__.__name__

    return f"{path_prefix}{chain_prefix}"


def get_streaming_callbacks(chain: Chain) -> list[Callable]:
    """Get streaming callbacks for a LangChain instance.

    Note: This function might not support all LangChain
    chain and agent types. Please open an issue on GitHub to
    request support for a specific type.

    Args:
        chain: A LangChain instance.
    """
    callbacks = []

    if "source_documents" in chain.output_keys:
        callbacks.append(SourceDocumentsStreamingCallbackHandler())

    if len(set(chain.output_keys) - {"source_documents"}) > 1:
        logger.warning(
            f"""multiple output keys found: {set(chain.output_keys) - {'source_documents'}}.

        Only the first output key will be used for streaming tokens. For more complex API logic, define the endpoint function manually.
        """
        )

    if isinstance(chain, AgentExecutor):
        callbacks.append(FinalTokenStreamingCallbackHandler())
    else:
        callbacks.extend(
            [
                TokenStreamingCallbackHandler(output_key=chain.output_keys[0]),
            ]
        )

    return callbacks


def get_websocket_callbacks(chain: Chain, websocket: WebSocket) -> list[Callable]:
    """Get websocket callbacks for a LangChain instance.

    Note: This function might not support all LangChain
    chain and agent types. Please open an issue on GitHub to
    request support for a specific type.

    Args:
        chain: A LangChain instance.
        websocket: A WebSocket instance.
    """
    callbacks = []

    if "source_documents" in chain.output_keys:
        callbacks.append(SourceDocumentsWebSocketCallbackHandler(websocket=websocket))

    if len(set(chain.output_keys) - {"source_documents"}) > 1:
        logger.warning(
            f"""multiple output keys found: {set(chain.output_keys) - {'source_documents'}}.

        Only the first output key will be used for sending tokens. For more complex websocket logic, define the endpoint function manually.
        """
        )

    if isinstance(chain, AgentExecutor):
        callbacks.append(FinalTokenWebSocketCallbackHandler(websocket=websocket))
    else:
        callbacks.extend(
            [
                TokenWebSocketCallbackHandler(
                    output_key=chain.output_keys[0], websocket=websocket
                ),
            ]
        )

    return callbacks



from enum import Enum
from typing import Any

from fastapi import status
from sse_starlette.sse import EventSourceResponse
from starlette.types import Send

from lanarky.events import Events, ServerSentEvent, ensure_bytes
from lanarky.logging import logger


class HTTPStatusDetail(str, Enum):
    INTERNAL_SERVER_ERROR = "Internal Server Error"


class StreamingResponse(EventSourceResponse):
    """`Response` class for streaming server-sent events.

    Follows the
    [EventSource protocol](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events#interfaces)
    """

    def __init__(
        self,
        content: Any = iter(()),
        *args: Any,
        **kwargs: dict[str, Any],
    ) -> None:
        """Constructor method.

        Args:
            content: The content to stream.
        """
        super().__init__(content=content, *args, **kwargs)

    async def stream_response(self, send: Send) -> None:
        """Streams data chunks to client by iterating over `content`.

        If an exception occurs while iterating over `content`, an
        internal server error is sent to the client.

        Args:
            send: The send function from the ASGI framework.
        """
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )

        try:
            async for data in self.body_iterator:
                chunk = ensure_bytes(data, self.sep)
                logger.debug(f"chunk: {chunk.decode()}")
                await send(
                    {"type": "http.response.body", "body": chunk, "more_body": True}
                )
        except Exception as e:
            logger.error(f"body iterator error: {e}")
            chunk = ServerSentEvent(
                data=dict(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=HTTPStatusDetail.INTERNAL_SERVER_ERROR,
                ),
                event=Events.ERROR,
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": ensure_bytes(chunk, None),
                    "more_body": True,
                }
            )

        await send({"type": "http.response.body", "body": b"", "more_body": False})



import re
from typing import Any, Awaitable, Callable

from fastapi import Depends
from pydantic import BaseModel, create_model
from starlette.routing import compile_path

from lanarky.events import Events
from lanarky.logging import logger
from lanarky.utils import model_dump, model_fields
from lanarky.websockets import WebSocket, WebsocketSession

from .resources import ChatCompletion, ChatCompletionResource, Message, OpenAIResource
from .responses import HTTPStatusDetail, StreamingResponse, status


def build_factory_api_endpoint(
    path: str, endpoint: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """Build a factory endpoint for API routes.

    Args:
        path: The path for the route.
        endpoint: openai resource factory function.
    """
    resource = compile_openai_resource_factory(endpoint)

    # index 1 of `compile_path` contains path_format output
    model_prefix = compile_model_prefix(compile_path(path)[1], resource)
    request_model = create_request_model(resource, model_prefix)

    async def factory_endpoint(
        request: request_model, resource: OpenAIResource = Depends(endpoint)
    ):
        return StreamingResponse(resource=resource, **model_dump(request))

    return factory_endpoint


def build_factory_websocket_endpoint(
    path: str, endpoint: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """Build a factory endpoint for WebSocket routes.

    Args:
        path: The path for the route.
        endpoint: openai resource factory function.
    """
    resource = compile_openai_resource_factory(endpoint)

    # index 1 of `compile_path` contains path_format output
    model_prefix = compile_model_prefix(compile_path(path)[1], resource)
    request_model = create_request_model(resource, model_prefix)

    async def factory_endpoint(
        websocket: WebSocket, resource: OpenAIResource = Depends(endpoint)
    ):
        async with WebsocketSession().connect(websocket) as session:
            async for data in session:
                try:
                    async for chunk in resource.stream_response(
                        **model_dump(request_model(**data))
                    ):
                        await websocket.send_json(
                            dict(
                                data=chunk,
                                event=Events.COMPLETION,
                            )
                        )
                except Exception as e:
                    logger.error(f"openai error: {e}")
                    await websocket.send_json(
                        dict(
                            data=dict(
                                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                                detail=HTTPStatusDetail.INTERNAL_SERVER_ERROR,
                            ),
                            event=Events.ERROR,
                        )
                    )
                await websocket.send_json(dict(data="", event=Events.END))

    return factory_endpoint


def compile_openai_resource_factory(endpoint: Callable[..., Any]) -> OpenAIResource:
    """Compile an OpenAI resource factory function.

    Args:
        endpoint: openai resource factory function.

    Returns:
        An OpenAIResource instance.
    """
    try:
        resource = endpoint()
    except TypeError:
        raise TypeError("set default values for all factory endpoint parameters")

    if not isinstance(resource, OpenAIResource):
        raise TypeError("factory endpoint must return a LanarkyOpenAIResource instance")
    return resource


def compile_model_prefix(path: str, resource: OpenAIResource) -> str:
    """Compile a prefix for pydantic models.

    Args:
        path: The path for the route.
        resource: An OpenAIResource instance.
    """
    # Remove placeholders like '{item}' using regex
    path_wo_params = re.sub(r"\{.*?\}", "", path)
    path_prefix = "".join([part.capitalize() for part in path_wo_params.split("/")])

    resource_prefix = resource.__class__.__name__

    return f"{path_prefix}{resource_prefix}"


def create_request_model(
    resource: ChatCompletionResource, prefix: str = ""
) -> BaseModel:
    """Create a pydantic model for incoming requests.

    Note: Support limited to ChatCompletion resource.

    Args:
        resource: An OpenAIResource instance.
        prefix: A prefix for the model name.
    """
    if not isinstance(resource, ChatCompletionResource):
        raise TypeError("resource must be a ChatCompletion instance")

    request_fields = {"messages": (list[Message], ...)}

    prefix = prefix or resource.__class__.__name__

    return create_model(f"{prefix}Request", **request_fields)


def create_response_model(
    resource: ChatCompletionResource, prefix: str = None
) -> BaseModel:
    """Create a pydantic model for responses.

    Note: Support limited to ChatCompletion resource.

    Args:
        resource: An OpenAIResource instance.
        prefix: A prefix for the model name.
    """
    if not isinstance(resource, ChatCompletionResource):
        raise TypeError("resource must be a ChatCompletion instance")

    response_fields = {
        k: (v.annotation, ...) for k, v in model_fields(ChatCompletion).items()
    }

    prefix = prefix or resource.__class__.__name__

    return create_model(f"{prefix}Response", **response_fields)



from enum import Enum

from sse_starlette.sse import ServerSentEvent as ServerSentEvent
from sse_starlette.sse import ensure_bytes as ensure_bytes


class Events(str, Enum):
    COMPLETION = "completion"
    ERROR = "error"
    END = "end"



from typing import Any, Callable, Optional

from fastapi import params
from langchain.chains.base import Chain

from lanarky.adapters.langchain.utils import create_request_model, create_response_model
from lanarky.utils import model_dump


def Depends(
    dependency: Optional[Callable[..., Any]],
    *,
    dependency_kwargs: dict[str, Any] = {},
    use_cache: bool = True
) -> params.Depends:
    """Dependency injection for LangChain.

    Args:
        dependency: a "dependable" chain factory callable.
        dependency_kwargs: kwargs to pass to chain dependency.
        use_cache: use_cache parameter of `fastapi.Depends`.
    """
    try:
        chain = dependency()
    except TypeError:
        raise TypeError("set default values for all dependency parameters")

    if not isinstance(chain, Chain):
        raise TypeError("dependency must return a Chain instance")

    request_model = create_request_model(chain)
    response_model = create_response_model(chain)

    async def chain_dependency(
        request: request_model,
        chain: Chain = params.Depends(dependency, use_cache=use_cache),
    ) -> response_model:
        return await chain.acall(inputs=model_dump(request), **dependency_kwargs)

    return params.Depends(chain_dependency, use_cache=use_cache)



from contextlib import asynccontextmanager
from enum import Enum
from typing import Generator

from fastapi.websockets import WebSocket, WebSocketDisconnect

from lanarky.logging import logger


class DataMode(str, Enum):
    JSON = "json"
    TEXT = "text"
    BYTES = "bytes"


class WebsocketSession:
    """Class to handle websocket connections.

    Supports 3 data modes: JSON, TEXT, and BYTES.

    To know more about WebSockets, read the
    [FastAPI documentation](https://fastapi.tiangolo.com/advanced/websockets/).
    """

    @asynccontextmanager
    async def connect(
        self, websocket: WebSocket, mode: DataMode = DataMode.JSON
    ) -> Generator:
        """Connect to a websocket and yield data from it.

        Args:
            websocket: The websocket to connect to.
            mode: The data mode to use. Defaults to DataMode.JSON.

        Yields:
            Any: data from client side.
        """
        await websocket.accept()
        try:
            if mode == DataMode.JSON:
                yield self.iter_json(websocket)
            elif mode == DataMode.TEXT:
                yield self.iter_text(websocket)
            elif mode == DataMode.BYTES:
                yield self.iter_bytes(websocket)
            else:
                raise ValueError(f"Invalid DataMode: {mode}")
        except WebSocketDisconnect:
            logger.info("Websocket disconnected")

    async def iter_text(self, websocket: WebSocket):
        while True:
            data = await websocket.receive_text()
            yield data

    async def iter_bytes(self, websocket: WebSocket):
        while True:
            data = await websocket.receive_bytes()
            yield data

    async def iter_json(self, websocket: WebSocket):
        while True:
            data = await websocket.receive_json()
            yield data



from abc import abstractmethod
from typing import Any, Generator

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from pydantic import BaseModel, Field

from lanarky.utils import model_dump


class Message(BaseModel):
    role: str = Field(pattern=r"^(user|assistant)$")
    content: str


class SystemMessage(BaseModel):
    role: str = "system"
    content: str


class OpenAIResource:
    """Base class for OpenAI resources."""

    def __init__(self, client: AsyncOpenAI = None):
        self._client = client or AsyncOpenAI()

    @abstractmethod
    async def stream_response(
        self, *args: Any, **kwargs: dict[str, Any]
    ) -> Generator[str, None, None]:
        ...


class ChatCompletionResource(OpenAIResource):
    """OpenAIResource class for chat completions."""

    def __init__(
        self,
        *,
        client: AsyncOpenAI = None,
        model: str = "gpt-3.5-turbo",
        stream: bool = False,
        system: str = None,
        **create_kwargs: dict[str, Any],
    ):
        """Constructor method.

        Args:
            client: An AsyncOpenAI instance.
            model: The model to use for completions.
            stream: Whether to stream completions.
            system: A system message to prepend to the messages.
            **create_kwargs: Keyword arguments to pass to the `chat.completions.create` method.
        """
        super().__init__(client=client)

        self.model = model
        self.stream = stream
        self.system = SystemMessage(content=system) if system else None
        self.create_kwargs = create_kwargs

    async def stream_response(self, messages: list[dict]) -> Generator[str, None, None]:
        """Stream chat completions.

        If `stream` attribute is False, the generator will yield only one completion.
        Otherwise, it will yield chunk completions.

        Args:
            messages: A list of messages to use for the completion.
                message format: {"role": "user", "content": "Hello, world!"}
        """
        messages = self._prepare_messages(messages)
        data = await self._client.chat.completions.create(
            messages=messages,
            model=self.model,
            stream=self.stream,
            **self.create_kwargs,
        )

        if self.stream:
            async for chunk in data:
                if not isinstance(chunk, ChatCompletionChunk):
                    raise TypeError(f"Unexpected data type: {type(data)}")
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        else:
            if not isinstance(data, ChatCompletion):
                raise TypeError(f"Unexpected data type: {type(data)}")
            yield data.choices[0].message.content

    async def __call__(self, messages: list[dict]) -> ChatCompletion:
        """Create a chat completion.

        Args:
            messages: A list of messages to use for the completion.
                message format: {"role": "user", "content": "Hello, world!"}

        Returns:
            A ChatCompletion instance.
        """
        messages = self._prepare_messages(messages)
        return await self._client.chat.completions.create(
            messages=messages,
            model=self.model,
            **self.create_kwargs,
        )

    def _prepare_messages(self, messages: list[dict]) -> list[dict]:
        if self.system is not None:
            messages = [model_dump(self.system)] + messages
        return messages
